/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file Client.h
 * \brief
 * MQTT client abstraction classes.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_CLIENT_H
#define OBLO_MESSENGER_CLIENT_H

#include "messenger/Network.h"
#include "messenger/NetworkProxy.h"
#include "messenger/Logging.h"

#include <map>

namespace Mqtt {

class DLLExport CAsyncClient: public Mqtt::CClientImpl,
    public Mqtt::INetworkProxyListener {
public:
    /**
     * @fn CAsyncClient(const std::string& _szClientId,
                 const std::string& _szServerUri,
                 IDeliveryListener& _DeliveryListener,
                 Messenger::ILogListener* _pLogListener=NULL);
     * @brief Constructor.
     * @param _szClientId MQTT Client ID.
     * @param _szServerUri MQTT Server URI.
     * @param _DeliveryListener Listener for arrived messages.
     * @param _pLogListener Log listener.
     */
    CAsyncClient(const std::string& _szClientId,
                 const std::string& _szServerUri,
                 IDeliveryListener* _pDeliveryListener,
                 Messenger::ILogListener* _pLogListener = NULL);
    /**
     * @fn virtual ~CAsyncClient();
     * @brief Destructor.
     */
    virtual ~CAsyncClient();
    /**
     * @brief ActivateThreadPool
     * Dispatch received messages from threads taken from a thread pool.
     * By default, thread pool is not activated and all received messages will be dispatched from a single thread.
     * @param _nMinThreads Set Minimum threads to run in the pool
     * @param _nMaxRecvThreads Set Maximum allowed threads to use for dispatching
     */
    void ActivateThreadPool(int _nMinThreads, int _nMaxThreads);
    /**
     * @brief DeactivateThreadPool Stop dispatching from thread pool, if previouslly configured
     */
    void DeactivateThreadPool();
    /**
     * @brief Setter for MQTT server URI. Must be called before Connect().
     * @param _szServerURI URI to use on connect.
     */
    void SetServerURI(const std::string& _szServerURI);
    /**
     * @brief GetServerUri
     * @return Configured server URI.
     */
    std::string GetServerUri() const;
    /**
     * @fn void SetUsername(const std::string& _szUsername);
     * @brief Setter for username. Must be called before Connect().
     * @param _szUsername Username to set.
     */
    void SetUsername(const std::string& _szUsername);
    /**
     * @fn void SetPassword(const std::string& _szPassword);
     * @brief Setter for password. Must be called before Connect().
     * @param _szPassword Password to set.
     */
    void SetPassword(const std::string& _szPassword);
    /**
     * @fn void SetCleanSession(bool _bCleanSession);
     * @brief Setter for clean session value. Must be called before Connect().
     * @param _bCleanSession Clean session on disconnect.
     */
    void SetCleanSession(bool _bCleanSession);
    /**
     * @fn bool IsCleanSession() const;
     * @brief Indicates whether session is cleand on disconnect.
     * @return Cleand session value.
     */
    bool IsCleanSession() const;
    /**
     * @fn void SetKeepAliveInterval(int _nInterval);
     * @brief Setter for keep alive interval. Must be called before Connect().
     * @param _nInterval Interval within ping to server will be set.
     */
    void SetKeepAliveInterval(int _nInterval);
    /**
     * @fn void SetConnectionTimeoutInterval(int _nInterval);
     * @brief Setter for connection timeout interval. Must be called before Connect().
     * @param _nInterval Interval to wait connection to complete.
     */
    void SetConnectionTimeoutInterval(int _nInterval);
    /**
     * @fn void SetWillConfiguration(const std::string& _szTopic, const Mqtt::CMessage::Ptr _pMsg);
     * @brief Configure message to be sent after client is disconnected.
     * @param _szTopic MQTT topic.
     * @param _pMsg MQTT message.
     */
    void SetWillConfiguration(const std::string& _szTopic,
                              const Mqtt::CMessage::Ptr _pMsg);
    /**
     * @fn void DisableLastWillMessage();
     * @brief Disables MQTT last will message.
     */
    void DisableLastWillMessage();
    /**
     * @fn void SetSecureConfiguration(const std::string& _szTrustStore, const std::string& _szKeyStore="", const std::string& _szPrivateKey="", const std::string& _szPrivateKeyPassword="");
     * @brief Set secure configuration. Connect must be called again, to apply this configuration.
     * @param _szTrustStore Secure connection trust store.
     * @param _szKeyStore Secure connection key store.
     * @param _szPrivateKey Secure connection private key.
     * @param _szPrivateKeyPassword Secure connection pasword for a provate key.
     */
    void SetSecureConfiguration(const std::string& _szTrustStore,
                                const std::string& _szKeyStore = "", const std::string& _szPrivateKey = "",
                                const std::string& _szPrivateKeyPassword = "");
    /**
     * @fn void Connect();
     * @brief Connects to messaging server.
     * @throw Poco::RuntimeException
     * @throw Poco::InvalidArgumentException
     */
    virtual void Connect();
    /**
     * @fn void Disconnect();
     * @brief Disconnects from a messaging server, if it was previusly connected.
     * @throw Poco::RuntimeException
     */
    virtual void Disconnect();
    /**
     * @fn void Publish(const std::string& _szTopic, const CMessage::Ptr _pMsg);
     * @brief Asynchronous publishing.
     * @param _szTopic Topic.
     * @param _pMsg MQTT message.
     */
    void Publish(const std::string& _szTopic, const CMessage::Ptr _pMsg);
    /**
     * @fn void RegisterConnectionListener(Messenger::IConnectionListener* _pNetworkStatusListener);
     * @brief Function used to register for events about network status.
     * @param _pNetworkStatusListener Network connection listener.
     */
    void RegisterConnectionListener(Messenger::IConnectionListener*
                                    _pConnectionListener);
    /**
     * @fn void UnregisterConnectionListener();
     * @brief Unregisters network connection listener.
     */
    void UnregisterConnectionListener();
protected:
    /**
     * @fn virtual void OnMessageArrived(const std::string &_szTopic, const CMessage::Ptr _pMsg);
     * @brief Method called when new message is arrived. It is redirected to delivery listener.
     * @param _szTopic Topic.
     * @param _pMsg MQTT message.
     */
    virtual void OnMessageArrived(const std::string& _szTopic,
                                  const CMessage::Ptr _pMsg);
    /**
     * @fn virtual void OnPublishFailed(int _nCode, const std::string &_szTopic, const CMessage::Ptr _pMsg);
     * @brief Used to report publish failure to delivery listener.
     * @param _nCode MQTT error code.
     * @param _szTopic Topic.
     * @param _pMsg MQTT message.
     */
    virtual void OnPublishFailed(int _nCode, const std::string& _szTopic,
                                 const CMessage::Ptr _pMsg);
    /**
     * @fn virtual void OnConnectionLost(const std::string _szCause);
     * @brief Redirects connection lost to connection listener.
     * @param _szCause Connection lost cause.
     */
    virtual void OnConnectionLost(const std::string& _szCause);
    /**
     * @fn virtual void OnPublishRequired(const std::string &_szTopic, const CMessage::Ptr _pMsg);
     * @brief Publishes messages.
     * @param _szTopic Topic.
     * @param _pMsg MQTT message.
     */
    virtual void OnPublishRequired(const std::string& _szTopic,
                                   const CMessage::Ptr _pMsg);

protected:
    Mqtt::IDeliveryListener* m_pDeliveryListener;
    Messenger::IConnectionListener* m_pConnectionListener;
    Messenger::ILogListener* m_pLogListener;
private:
    Mqtt::CConnectConfig::Ptr m_pConnConfig;
    Mqtt::CSslConfig::Ptr m_pSslConfig;
    Mqtt::CWillConfig::Ptr m_pWillConfig;

    Mqtt::CAsyncNetworkProxy m_NetworkProxy;
};

class DLLExport CBuffClient: public CAsyncClient {
public:
    /**
     * @fn CBuffClient(const std::string& _szClientId,
                   const std::string& _szServerUri,
                   IDeliveryListener& _DeliveryListener,
                   Messenger::ILogListener* _pLogListener=NULL);
     * @brief Constructor
     * @param _szClientId Client ID.
     * @param _szServerUri Server URI.
     * @param _DeliveryListener Delivery listener.
     * @param _pLogListener Log listener.
     * @param _nMaxQueueSize Maximum number of publish messages stored to the buffer.
     */
    CBuffClient(const std::string& _szClientId,
                const std::string& _szServerUri,
                IDeliveryListener& _DeliveryListener,
                Messenger::ILogListener* _pLogListener = NULL,
                unsigned int _nMaxQueueSize = 100);
    /**
     * @fn virtual void Connect();
     * @brief If basic connect succeeds it recovers subscriptions and failed publish messages.
     * @throws Poco::RuntimeException
     * @throws Poco::InvalidArgumentException
     */
    virtual void Connect();
    void ConnectBasic();
    /**
     * @fn void Subscribe(const std::string& _szTopic, int _nQos);
     * @brief If client is not connected, it stores subscriptions and will try to complete the procedure on next connect.
     * @param _szTopic Topic to subscribe.
     * @param _nQos QoS level.
     */
    void Subscribe(const std::string& _szTopic, int _nQos);
    /**
     * @fn void Subscribe(const TopicCollection& _TopicFilters, const QosCollection& _Qos, bool _bStore=true);
     * @brief If client is not connected, it stores subscriptions and will try to complete the procedure on next connect.
     * @param _TopicFilters Topics to subscribe.
     * @param _Qos QoS levels.
     * @throws Poco::InvalidArgumentException
     */
    void Subscribe(const TopicCollection& _TopicFilters, const QosCollection& _Qos);
    /**
     * @fn virtual void Unsubscribe(const std::string& _szTopic);
     * @brief Does basic unsubscribe and removes stored subscription.
     * @param _szTopic Topic.
     */
    virtual void Unsubscribe(const std::string& _szTopic);
    /**
     * @fn virtual void Unsubscribe(const TopicCollection& _TopicFilters);
     * @brief Does basic unsubscribe and removes stored subscriptions.
     * @param _TopicFilters Topics.
     */
    virtual void Unsubscribe(const TopicCollection& _TopicFilters);
    /**
     * @brief Control subscriptions if a session is present on the broker. Session present flag will be checked and if it is false the subscription will be applied. Otherwise, subscription will be skipped.
     * @param _bIgnore Indication how to handle the session present flag.
     */
    void IgnoreSessionPresent(bool _bIgnore);
private:
    typedef std::pair<std::string, int> SubElement;
    typedef std::deque<SubElement> SubscriptionBuffer;

    /**
     * @fn virtual void OnPublishFailed(int _nCode, const std::string& _szTopic, const CMessage::Ptr _pMsg);
     * @brief Stores QoS1 and QoS2 messages and tries to publish them on next connect. QoS messages are handled by @see CAsyncClient
     * @param _nCode Error code.
     * @param _szTopic Topic.
     * @param _pMsg MQTT message
     */
    virtual void OnPublishFailed(int _nCode, const std::string& _szTopic,
                                 const CMessage::Ptr _pMsg);
    /**
     * @fn virtual void StoreSubscription(const TopicCollection& _Topics, const QosCollection& _QoS);
     * @brief Stores subscriptions for next connect.
     * @param _Topics Topic collection.
     * @param _QoS QoS collection.
     */
    virtual void StoreSubscription(const TopicCollection& _Topics,
                                   const QosCollection& _QoS);
    /**
     * @fn virtual void StoreUnsubscription(const std::string& _szTopic, int _nQoS);
     * @brief Stores subscriptions for next connect.
     * @param _Topics Topic collection.
     */
    virtual void StoreUnsubscription(const TopicCollection& _Topics);
    /**
     * @fn void RemoveSubscription(const TopicCollection& _Topics);
     * @brief Removes subscriptions that are stored for next connect.
     * @param _Topics Topic collection.
     */
    void RemoveSubscription(const TopicCollection& _Topics);
    /**
     * @fn void RemoveUnsubscription(const TopicCollection& _Topics);
     * @brief Removes unsubscriptions that are stored for next connect.
     * @param _Topics Topic collection.
     */
    void RemoveUnsubscription(const TopicCollection& _Topics);
    /**
     * @fn void Remove(SubscriptionBuffer& _Buffer, const TopicCollection& _Topics);
     * @brief Removes subscriptions that are stored for next connect.
     * @param _Buffer Subscription buffer.
     * @param _Topics Topic collection.
     */
    void Remove(SubscriptionBuffer& _Buffer, const TopicCollection& _Topics);
    /**
     * @fn void PublishStore();
     * @brief Tries to publish stored messages.
     */
    void PublishStore();
    /**
     * @fn void SubscribeStore();
     * @brief Tries to subscribe to stored topics.
     */
    void SubscribeStore();
    /**
     * @fn void UnsubscribeStore();
     * @brief Tries to unsubscribe from stored topics.
     */
    void UnsubscribeStore();

    SubscriptionBuffer m_Subscriptions;
    Poco::Mutex m_SubscribeMutex;

    SubscriptionBuffer m_Unsubscriptions;
    Poco::Mutex m_UnsubscribeMutex;

    typedef std::queue<CPublishHolder::Ptr> PublicationBuffer;
    PublicationBuffer m_Publications;
    Poco::Mutex m_PublishMutex;

    const unsigned int m_nMaxQueueSize;
    bool m_bIgnoreSessionPresent;
};

}

#endif
