/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file Messages.h
 * \brief
 * Classes for different types of messages arriving from the network.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_MESSEGES_H
#define OBLO_MESSENGER_MESSEGES_H

#include "Poco/DynamicStruct.h"
#include "Poco/SynchronizedObject.h"
#include "Poco/Random.h"

#include "messenger/Common.h"

namespace Messenger {

typedef enum {
    PROTOCOL_OBLO_1 = 0,
    PROTOCOL_WISE_1
} ProtocolVersion;

class CResponseMessage;
class CRequestMessage;

#define MAKE_REQ_DELEGATE(tp, ms, mf, o) \
    Messenger::CRequestDelegate::CreateDelegate<tp, & tp :: ms, & tp :: mf>(o)
#define MAKE_AUTO_DEL_REQ_DELEGATE(tp, ms, mf, o) \
    Messenger::CRequestDelegate::CreateDelegate<tp, & tp :: ms, & tp :: mf>(o, true)
/**
 * @class CRequestDelegate
 * @brief Delegate used to report new response messages.
 */
class DLLExport CRequestDelegate {
public:
    typedef Poco::SharedPtr<CRequestDelegate> Ptr;
    /**
     * @fn CRequestDelegate();
     * @brief Default constructor.
     */
    CRequestDelegate();

    template <class T, void (T::*TSuccessMethod)(unsigned int _nRequestId, const CResponseMessage& _Response), void (T::*TFailMethod)(unsigned int _nRequestId)>
    /**
     * @fn static CRequestDelegate::Ptr CreateDelegate(T* _pObject)
     * @param _pObject Object pointer.
     * @param _bAutoDelete Delete _pObject after a delegation is completed.
     * @return Pointer to a request delegate.
     */
    static CRequestDelegate::Ptr CreateDelegate(T* _pObject, bool _bAutoDelete = false) {
        CRequestDelegate::Ptr d(new CRequestDelegate());
        d->m_pObject = _pObject;
        d->m_pStubSucc = &MethodStubSucc<T, TSuccessMethod>;
        d->m_pStubFail = &MethodStubFail<T, TFailMethod>;
        d->m_bAutoDelete = _bAutoDelete;
        return d;
    }
    /**
     * @fn void ReportSuccess(unsigned int _nRequestId, const CResponseMessage& _Response) const;
     * @param _nRequestId Request ID.
     * @param _Response A response reference.
     */
    void ReportSuccess(unsigned int _nRequestId, const CResponseMessage& _Response) const;
    /**
     * @fn void ReportFailure(unsigned int _nRequestId, const CResponseMessage& _Response) const;
     * @param _nRequestId Request ID.
     */
    void ReportFailure(unsigned int _nRequestId) const;

private:
    typedef void (*stub_type_succ)(void* m_pObject, unsigned int _nRequestId, const CResponseMessage& _Response, bool _bAutoDel);
    typedef void (*stub_type_fail)(void* m_pObject, unsigned int _nRequestId, bool _bAutoDel);

    void* m_pObject;
    stub_type_succ m_pStubSucc;
    stub_type_fail m_pStubFail;
    bool m_bAutoDelete;

    template <class T, void (T::*TMethodSucc)(unsigned int _nRequestId, const CResponseMessage& _pResponse)>
    static void MethodStubSucc(void* _pObject, unsigned int _nRequestId, const CResponseMessage& _pResponse, bool _bAutoDel) {
        T* p = static_cast<T*>(_pObject);
        (p->*TMethodSucc)(_nRequestId, _pResponse);

        if (_bAutoDel) {
            SAFE_DELETE_O(p);
        }
    }
    template <class T, void (T::*TMethodFail)(unsigned int _nRequestId)>
    static void MethodStubFail(void* _pObject, unsigned int _nRequestId, bool _bAutoDel) {
        T* p = static_cast<T*>(_pObject);
        (p->*TMethodFail)(_nRequestId);

        if (_bAutoDel) {
            SAFE_DELETE_O(p);
        }
    }

};

/**
 * @class CMessageImpl
 * @brief Implementation of basic functionalities for messages.
 */
class DLLExport CMessageImpl {
public:
    static const Poco::Timestamp::TimeVal EXPIRATION_NOT_SET;
    typedef Poco::SharedPtr<CMessageImpl> Ptr;
    /**
     * @fn CMessageImpl();
     * @brief Default constructor.
     */
    CMessageImpl();
    /**
     * @fn virtual ~CMessageImpl();
     * @brief Destructor.
     */
    virtual ~CMessageImpl() {}
    /**
     * @fn CMessageImpl(const CMessageImpl& _Message);
     * @brief Copy-constructor.
     * @param _Message Message.
     */
    CMessageImpl(const CMessageImpl& _Message);
    /**
     * @fn CMessageImpl& operator =(const CMessageImpl& _Msg);
     * @brief Assign operator.
     * @param _Msg Message object.
     * @return Reference to an object.
     */
    CMessageImpl& operator =(const CMessageImpl& _Msg);
    /**
     * @fn unsigned int GetId() const;
     * @return
     */
    unsigned int GetId() const;
    /**
     * @fn const Poco::DynamicStruct::Ptr GetParams() const;
     * @return Pointer to params, which can be NULL if params is not set.
     */
    const Poco::DynamicStruct::Ptr GetParams() const;
    /**
     * @fn Poco::DynamicStruct& GetParamsHolder() const;
     * @return Reference to the params struct inside the message. Can be filled directlly.
     */
    Poco::DynamicStruct& GetParamsHolder();
    /**
     * @fn void SetParams(Poco::DynamicStruct::Ptr _pParams);
     * @param _pParams
     */
    void SetParams(Poco::DynamicStruct::Ptr _pParams);
    /**
     * GetTimestamp Event time getter
     * @return Timestamp when the event is created
     */
    Poco::Timestamp GetTimestamp() const;
    /**
     * @brief SetTimestamp time setter
     * @param _Timestamp Time when the event is created
     */
    void SetTimestamp(const Poco::Timestamp& _Timestamp);
    /**
     * @brief GetExpirationTime Get expiration time
     * @return Timestamp
     */
    Poco::Timestamp GetExpirationTime() const;
    /**
     * @brief IsExpirationSet Check if the expiration time is configured
     * @return Indication if the expiration is present in the event message
     */
    bool IsExpirationSet() const;
    /**
     * @fn virtual std::string ToJson(ProtocolVersion _eProtocol);
     * @param _eProtocol Protocol version
     * @return
     */
    virtual std::string ToJson(ProtocolVersion _eProtocol);
    /**
     * @fn virtual void ToJson(ProtocolVersion _eProtocol, std::string& _szOutput);
     * @param _eProtocol Protocol version
     * @param _szOutput Output string
     */
    virtual void ToJson(ProtocolVersion _eProtocol, std::string& _szOutput);
protected:
    friend class CMessenger;
    /**
     * @fn void SetId(unsigned int _nId);
     * @param _nId
     */
    void SetId(unsigned int _nId);
    /**
     * Setting timestamp, for internal use only.
     * @param timestamp to be set
     */
    void SetTimestamp(const std::string& _szTimestamp);
    /**
     * @fn virtual void InflateJson(Poco::DynamicStruct _Obj, ProtocolVersion _eProtocol) const=0;
     * @brief Abstract method used to inflate fields from a subclass into a JSON object.
     * @param _Obj
     * @param _eProtocol Protocol version
     */
    virtual void InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const = 0;
    /**
     * @fn static void Parse(CMessageImpl& _MessageToFill, const Poco::DynamicStruct& _MessageToParse);
     * @param _pMessageToFill Object to fill with data from parsing object.
     * @param _pMessageToParse Object to parse.
     * @param _eProtocol Protocol version.
     * @throw Poco::Exception
     */
    static void Parse(CMessageImpl& _MessageToFill, const Poco::DynamicStruct& _MessageToParse, ProtocolVersion _eProtocol);

private:
    static Poco::Timestamp FromEpochMilliseconds(Poco::Timestamp::TimeVal _nMilliseconds);

    static Poco::Random RANDOM_GENERATOR;

    unsigned int m_nId;
    Poco::DynamicStruct::Ptr m_pParams;
    Poco::Timestamp m_tTimestamp;
    Poco::Timestamp m_tExpirationTime;
    // Store message to dynamic var to be able to convert to JSON without additional copy
    Poco::Dynamic::Var m_Message;
};
/**
 * @class  CEventMessage
 * @brief Wrapper for messages about events arriving from objects included in communication.
 */
class DLLExport CEventMessage: public CMessageImpl {
public:
    typedef Poco::SharedPtr<CEventMessage> Ptr;
    /**
     * @fn CEventMessage();
     * @brief Default constructor.
     */
    CEventMessage();
    /**
     * @fn CEventMessage(const CEventMessage& _Event);
     * @brief Copy-constructor.
     * @param _Event Event object.
     */
    CEventMessage(const CEventMessage& _Event);
    /**
     * @fn CEventMessage& operator =(const CEventMessage& _Event);
     * @brief Assign operator.
     * @param _Event Event object.
     * @return Reference to an event object.
     */
    CEventMessage& operator =(const CEventMessage& _Event);
    /**
     * @fn const std::string& GetName() const;
     * @return
     */
    std::string GetName() const;
    /**
     * @fn void SetName(const std::string& _szName);
     * @param _szName
     */
    void SetName(const std::string& _szName);
    /**
     * Get timestamp in ISO8601.
     * @return Timestamp in ISO8601 format
     */
    std::string GetTimestampString() const;
    /**
     * @brief static CEventMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
     * @param _Message Object to parse.
     * @param _eProtocol Protocol version
     * @return Pointer to a message.
     * @throw Poco::DataFormatException
     */
    static CEventMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
protected:
    /**
     * @fn virtual void InflateJson(Poco::DynamicStruct _Obj) const=0;
     * @brief Method to inflate class members to JSON object.
     * @param _Obj
     * @param _eProtocol Protocol version
     */
    virtual void InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const;
private:
    std::string m_szName;
};
/**
 * @class  CLastwillMessage
 * @brief Wrapper for messages about lastwill arriving from objects included in communication.
 */
class DLLExport CLastwillMessage: public CMessageImpl  {
public:
    typedef Poco::SharedPtr<CLastwillMessage> Ptr;
    /**
     * @fn CLastwillMessage();
     * @brief Default constructor.
     */
    CLastwillMessage();
    /**
     * @fn CLastwillMessage(const CLastwillMessage& _Obj);
     * @brief Copy-constructor.
     * @param _Obj Object.
     */
    CLastwillMessage(const CLastwillMessage& _Obj);
    /**
     * @fn CLastwillMessage& operator =(const CLastwillMessage& _Obj);
     * @brief Assign operator.
     * @param _Obj Object.
     * @return Reference to a lastwill object.
     */
    CLastwillMessage& operator =(const CLastwillMessage& _Obj);
    /**
     * @fn bool IsOnline() const;
     * @return
     */
    bool IsOnline() const;
    /**
     * @fn void SetStatus(bool _bOnline);
     * @param _bOnline
     */
    void SetStatus(bool _bOnline);
    /**
     * @fn static CLastwillMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
     * @param _Message Object to parse.
     * @param _eProtocol Protocol version
     * @return Pointer to a message.
     * @throw Poco::DataFormatException
     */
    static CLastwillMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
protected:
    /**
     * @fn virtual void InflateJson(Poco::DynamicStruct _Obj) const=0;
     * @brief Method to inflate class members to JSON object.
     * @param _Obj
     * @param _eProtocol Protocol version
     */
    virtual void InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const;
private:
    bool m_bOnline;
};
/**
 * @class  CResponseMessage
 * @brief Wrapper for response messages included in communication.
 */
class DLLExport CResponseMessage : public CMessageImpl {
    friend class CMessenger;
    friend class CUser;
public:
    typedef Poco::SharedPtr<CResponseMessage> Ptr;
    /**
     * @fn CResponseMessage();
     * @brief Default constructor.
     */
    CResponseMessage();
    /**
     * @fn CResponseMessage(const CResponseMessage& _Response);
     * @brief Copy-constructor.
     * @param _Response Response object.
     */
    CResponseMessage(const CResponseMessage& _Response);
    /**
     * @fn CResponseMessage& operator =(const CResponseMessage& _Response);
     * @brief Assign operator.
     * @param _Response Response object.
     * @return Reference to a response object.
     */
    CResponseMessage& operator =(const CResponseMessage& _Response);
    /**
     * @brief static CResponseMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
     * @param _Message Object to parse.
     * @param _eProtocol Protocol version
     * @return Pointer to a message.
     * @throw Poco::DataFormatException
     */
    static CResponseMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
    /**
     * @fn int GetCode() const;
     * @return Response code.
     */
    int GetCode() const;
    /**
     * @fn void SetCode(int _nCode);
     * @param _nCode Response code.
     */
    void SetCode(int _nCode);
    /**
     * @fn std::string GetMessage() const;
     * @return Response message.
     */
    std::string GetMessage() const;
    /**
     * @fn void SetMessage(const std::string& _szMessage);
     * @param _szMessage Response message.
     */
    void SetMessage(const std::string& _szMessage);

protected:
    /**
     * @fn virtual void InflateJson(Poco::DynamicStruct _Obj) const=0;
     * @brief Method to inflate class members to JSON object.
     * @param _Obj
     * @param _eProtocol Protocol version
     */
    virtual void InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const;
private:
    int m_nCode;
    std::string m_szMessage;
};

class ICompletionListener {
public:
    virtual void OnRequestCompleted(unsigned int _nReqId) = 0;
};
/**
 * @class  CRequestMessage
 * @brief Wrapper for request messages.
 */
class DLLExport CRequestMessage : public CMessageImpl {
public:
    typedef Poco::SharedPtr<CRequestMessage> Ptr;
    /**
     * @fn CRequestMessage();
     * @brief Default constructor.
     */
    CRequestMessage();
    /**
     * @fn CRequestMessage(const CRequestMessage& _Request);
     * @brief Copy-constructor.
     * @param _Request Request object.
     */
    CRequestMessage(const CRequestMessage& _Request);
    /**
     * @fn CRequestMessage& operator =(const CRequestMessage& _Request);
     * @brief Assign operator.
     * @param _Request Request object.
     * @return Reference to an request object.
     */
    CRequestMessage& operator =(const CRequestMessage& _Request);
    /**
     * @fn std::string GetSender() const;
     * @return
     */
    std::string GetSender() const;
    /**
     * @fn std::string GetType() const;
     * @return
     */
    std::string GetType() const;
    /**
     * @fn void SetType(const std::string& _szType);
     * @param _szType
     */
    void SetType(const std::string& _szType);
    /**
     * @fn std::string GetName() const;
     * @return
     */
    std::string GetName() const;
    /**
     * @fn void SetName(const std::string& _szName);
     * @param _szName
     */
    void SetName(const std::string& _szName);
    /**
     * @brief static CRequestMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
     * @param _Message Object to parse.
     * @param _eProtocol Protocol version
     * @return Pointer to a message.
     * @throw Poco::DataFormatException
     */
    static CRequestMessage::Ptr Parse(const Poco::DynamicStruct& _Message, ProtocolVersion _eProtocol);
protected:
    friend class CMessenger;
    friend class CUser;
    /**
     * @fn void SetSender(const std::string& _szSender);
     * @param _szSender
     */
    void SetSender(const std::string& _szSender);
    /**
     * @fn virtual void InflateJson(Poco::DynamicStruct _Obj) const=0;
     * @brief Method to inflate class members to JSON object.
     * @param _Obj
     * @param _eProtocol Protocol version
     */
    virtual void InflateJson(Poco::DynamicStruct& _Obj, ProtocolVersion _eProtocol) const;
    /**
     * @fn void RegisterCompletionListener(ICompletionListener* _pCompletionListener);
     * @brief Registers listener on information about request status.
     * @param _pCompletionListener Listener.
     */
    void RegisterCompletionListener(ICompletionListener* _pCompletionListener);
    /**
     * @fn virtual void ReportResponse(const CResponseMessage::Ptr _pResponse);
     * @param _pResponse
     */
    virtual void ReportResponse(const CResponseMessage::Ptr _pResponse);
    /**
     * @fn virtual void ReportFailure();
     */
    virtual void ReportFailure();
private:
    std::string m_szSender;
    std::string m_szType;
    std::string m_szName;
    ICompletionListener* m_pCompletitionListener;
};
/**
 * @class CSyncRequest
 * @brief Class used to handle synchronized requests inside a messenger.
 */
class CSyncRequest : public CRequestMessage {
public:
    typedef Poco::SharedPtr<CSyncRequest> Ptr;
    /**
     * @fn CSyncRequest(const CRequestMessage& _Request);
     * @brief Constructor that wrapps request message.
     * @param _Request Request message.
     */
    CSyncRequest(const CRequestMessage& _Request);
    /**
     * @fn const CResponseMessage::Ptr GetResponse() const;
     * @return Response pointer to the request. NULL can be return value, if response is not set.
     */
    const CResponseMessage::Ptr GetResponse() const;
protected:
    friend class CUser;
    /**
     * @fn void Wait(unsigned int _nTimeout) const;
     * @param _nTimeout Time to wait.
     * @throw Poco::TimeoutException
     */
    void Wait(unsigned int _nTimeout) const;
    /**
     * @fn virtual void ReportResponse(CResponseMessage::Ptr _pResponse);
     * @param _pResponse Response that arrived.
     */
    virtual void ReportResponse(const CResponseMessage::Ptr _pResponse);
    /**
     * @fn virtual void ReportFailure();
     */
    virtual void ReportFailure();
private:
    /**
     * @fn CSyncRequest();
     * @brief Unimplemented default constructor. Left as it is, because of synchronized object.
     * @see Poco::SynchronizedObject
     */
    CSyncRequest();
    /**
     * @fn CSyncRequest();
     * @brief Unimplemented copy constructor. Left as it is, because of synchronized object.
     * @see Poco::SynchronizedObject
     */
    CSyncRequest(const CSyncRequest& _Request);
    /**
     * @fn CSyncRequest& operator =(const CSyncRequest& _Request);
     * @brief Unimplemented assign operator. Left as it is, because of synchronized object.
     * @see Poco::SynchronizedObject
     */
    CSyncRequest& operator =(const CSyncRequest& _Request);

    CResponseMessage::Ptr m_pResponse;
    Poco::SynchronizedObject m_SyncObj;
};
/**
 * @class CAsyncRequest
 * @brief Class used to handle asynchronous requests inside a messenger.
 */
class CAsyncRequest : public CRequestMessage {
public:
    typedef Poco::SharedPtr<CAsyncRequest> Ptr;
    /**
     * @fn CAsyncRequest(const CRequestMessage& _Request, CRequestDelegate::Ptr _pDelegate);
     * @brief Constructor that wrapps request message.
     * @param _Request Request message.
     * @param _pDelegate Delegate to inform when a response arrives.
     */
    CAsyncRequest(const CRequestMessage& _Request, CRequestDelegate::Ptr _pDelegate);
protected:
    /**
     * @fn virtual void ReportResponse(CResponseMessage::Ptr _pResponse);
     * @param _pResponse Response that arrived.
     */
    virtual void ReportResponse(const CResponseMessage::Ptr _pResponse);
    /**
     * @fn virtual void ReportFailure();
     */
    virtual void ReportFailure();
private:
    const CRequestDelegate::Ptr m_pRequestDelegate;
};

}
#endif
