/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file Messenger.h
 * \brief
 * Implementation of messagining inside of OBLO system.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_MESSENGER_H
#define OBLO_MESSENGER_MESSENGER_H

#include "messenger/Common.h"
#include "messenger/Logging.h"
#include "messenger/Messages.h"
#include "messenger/Client.h"
#include "messenger/User.h"
#include "Poco/RWLock.h"

namespace Messenger {

/**
 * @interface IResponder
 * @brief Used to get responses to request messages.
 */
class IResponder {
public:
    /**
     * @fn virtual const CResponseMessage::Ptr Respond(const CRequestMessage::Ptr _pRequest) = 0;
     * @brief Function which will be called when new requests arrive.
     * @param _pRequest Request message.
     * @return Response to request message.
     */
    virtual const CResponseMessage::Ptr Respond(const CRequestMessage::Ptr _pRequest) = 0;
    /**
     * @fn virtual const CResponseMessage::Ptr FunctionalityRespond(const std::string& _szFunctionality,
            const CRequestMessage::Ptr _pRequest) = 0;
     * @brief Function which will be called when new requests arrive to a functionality.
     * @param _szFunctionality Identifier of a functionality.
     * @param _pRequest Request message.
     * @return Response to request message.
     */
    virtual const Messenger::CResponseMessage::Ptr FunctionalityRespond(const std::string& _szFunctionality,
            const Messenger::CRequestMessage::Ptr _pRequest) = 0;
};
/**
 * @class CLogger
 * @brief The logger implementation.
 */
class DLLExport CLogger: protected ILogListener {
public:
    /**
     * @fn CLogger(const std::string& _szIdentifier);
     * @brief Constructor
     * @param _szIdentifier Logger ID.
     */
    CLogger(const std::string& _szIdentifier);
    /**
     * @fn void RegisterLogListener(ILogListener* _pLogListener);
     * @brief Registers log listener.
     * @param _pLogListener Log listener interface.
     */
    void RegisterLogListener(ILogListener* _pLogListener);
    /**
     * @fn void UnregisterLogListener();
     * @brief Cleans a log listener.
     */
    void UnregisterLogListener();
protected:
    /**
     * @fn void OnTrace(const std::string& _Msg);
     * @brief Handles trace messages.
     * @param _Msg Message.
     */
    void OnTrace(const std::string& _Msg);
    /**
     * @fn void OnTrace(const std::string& _Msg);
     * @brief Handles debug messages.
     * @param _Msg Message.
     */
    void OnDebug(const std::string& _Msg);
    /**
     * @fn void OnTrace(const std::string& _Msg);
     * @brief Handles information messages.
     * @param _Msg Message.
     */
    void OnInformation(const std::string& _Msg);
    /**
     * @fn void OnTrace(const std::string& _Msg);
     * @brief Handles warning messages.
     * @param _Msg Message.
     */
    void OnWarning(const std::string& _Msg);
    /**
     * @fn void OnTrace(const std::string& _Msg);
     * @brief Handles error messages.
     * @param _Msg Message.
     */
    void OnError(const std::string& _Msg);
private:
    ILogListener* m_pLogListener;
    const std::string m_szId;
};
/**
 * @class CCleaner
 * @brief Request cleaner implementation.
 */
class CCleaner {
protected:
    /**
     * @class CCleanupElement
     * @brief Wrapper for a cleanup ellement.
     */
    class CCleanupElement {
    public:
        typedef Poco::SharedPtr<CCleanupElement> Ptr;
        /**
         * @fn CCleanupElement(unsigned int _nId, int _nTimeout);
         * @brief Constructior
         * @param _nId Element to clean.
         * @param _nTimeout Time when cleaning should be done.
         */
        CCleanupElement(unsigned int _nId, int _nTimeout);
        /**
         * @fn CCleanupElement(const CCleanupElement& _Elm);
         * @brief Copy constructor.
         * @param _Elm Element to copy from.
         */
        CCleanupElement(const CCleanupElement& _Elm);
        /**
         * @fn CCleanupElement& operator =(const CCleanupElement& _Elm);
         * @brief Assignment operator.
         * @param _Elm Element to assigne fields from.
         * @return
         */
        CCleanupElement& operator =(const CCleanupElement& _Elm);
        unsigned int m_nId;
        int m_nTimeout;
    };
    /**
     * @fn CCleaner();
     * @brief Default constructor.
     */
    CCleaner();
    /**
     * @fn CCleaner(const CCleaner& _Cleaner);
     * @brief Copy constructor.
     * @param _Cleaner Element to copy from.
     */
    CCleaner(const CCleaner& _Cleaner);
    /**
     * @fn void StopCleaning();
     * @brief Stop cleaning process.
     */
    void StopCleaning();
    /**
     * @fn void ScheduleCleanup(unsigned int _nId, unsigned int _nTime);
     * @brief Schedule cleaning for a request.
     * @param _nId ID.
     * @param _nTime Relatime time to clean request in seconds.
     */
    void ScheduleCleanup(unsigned int _nId, unsigned int _nTime);
    /**
     * @fn void OnCleanup(Poco::Timer& timer);
     * @brief Callback called when async requests should be cleaned.
     * @param timer
     */
    void OnCleanup(Poco::Timer& timer);
    /**
     * @fn void CleanupAll();
     * @brief Cleanup all pending requests.
     */
    void CleanupAll();
    /**
     * @fn void CleanRequest(unsigned int _nReqId);
     * @brief Removes a request from the queue.
     * @param _nReqId Request identifier.
     */
    virtual void CleanRequest(unsigned int _nReqId) = 0;
private:
    typedef std::list<CCleanupElement::Ptr> CleanUpList;
    CleanUpList m_CleanupList;
    Poco::Mutex m_CleanupMutex;
    Poco::Timer m_CleanupTimer;
};

class DLLExport CMessenger: public CLogger, public Mqtt::IDeliveryListener, protected CCleaner {
public:
    /**
     * @fn CMessenger(const std::string& _szTag,
               const std::string& _szUsername,
               const std::string& _szPassword,
               const std::string& _szUri);
     * @brief Constructor which creates and includes an object in OBLO messaging.
     * @param _szTag Tag to be used to identify as an user.
     * @param _szUsername Username used to authorize as an user.
     * @param _szPassword Password used to authorize as an user.
     * @param _szUri URI of a messaging server.
     * @param _eProtocol Version of the protocol.
     */
    CMessenger(const std::string& _szTag,
               const std::string& _szUsername = "",
               const std::string& _szPassword = "",
               const std::string& _szUri = "tcp://localhost:1883",
               ProtocolVersion _eProtocol = PROTOCOL_OBLO_1);
    /**
     * @fn ~CMessenger();
     * @brief Destructor.
     */
    virtual ~CMessenger();
    /**
     * @brief ActivateThreadPool
     * Dispatch received messages from threads taken from a thread pool.
     * By default, thread pool is not activated and all received messages will be dispatched from a single thread.
     * @param _nMinThreads Set Minimum threads to run in the pool
     * @param _nMaxRecvThreads Set Maximum allowed threads to use for dispatching
     */
    void ActivateThreadPool(int _nMinThreads, int _nMaxThreads);
    /**
     * @brief DeactivateThreadPool Stop dispatching from thread pool, if previouslly configured
     */
    void DeactivateThreadPool();
    /**
     * @fn void SetSecureConfiguration(const std::string& _szTrustStore, const std::string& _szKeyStore, const std::string& _szPrivateKey, const std::string& _szPrivateKeyPassword);
     * @brief Set secure configuration. Connect must be called again, to apply this configuration.
     * @param _szTrustStore Secure connection trust store.
     * @param _szKeyStore Secure connection key store.
     * @param _szPrivateKey Secure connection private key.
     * @param _szPrivateKeyPassword Secure connection pasword for a provate key.
     */
    void SetSecureConfiguration(const std::string& _szTrustStore, const std::string& _szKeyStore, const std::string& _szPrivateKey, const std::string& _szPrivateKeyPassword);
    /**
     * @fn void Connect();
     * @brief Connects to messaging server.
     * @throw Poco::RuntimeException
     */
    void Connect();
    /**
     * @brief Reconnects to messaging server. Combination of disconnect and connect force clean session
     * @throw Poco::RuntimeException
     */
    void Reconnect(bool _bCleanSession);
    /**
     * @fn void Disconnect();
     * @brief Disconnects from a messaging server, if it was previusly connected.
     * @throw Poco::RuntimeException
     */
    void Disconnect();
    /**
     * @brief IsConnected
     * @return true/false
     */
    bool IsConnected() const;
    /**
     * @brief Control subscriptions if a session is present on the broker. Session present flag will be checked and if it is false the subscription will be applied. Otherwise, subscription will be skipped.
     * @param _bIgnore Indication how to handle the session present flag.
     */
    void IgnoreSessionPresent(bool _bIgnore);
    /**
     * @fn void RegisterResponder(IResponder* _pResponder);
     * @brief Registers a responder for requests arriving to messenger.
     * @param _pResponder Responder interface.
     * @throw Poco::RuntimeException
     */
    void RegisterResponder(IResponder* _pResponder);
    /**
     * @fn void UnregisterResponder();
     * @brief Unregisters from previusly defined responder.
     * @throw Poco::RuntimeException
     */
    void UnregisterResponder();
    /**
     * @fn void ActivateFunctionality(const std::string& _szFunctionality);
     * @brief Registers for requests arriving on a functionality. Responder needs to be registered to receive corresponding messages.
     * @param _szFunctionality Identifier of a functionality.
     * @throw Poco::RuntimeException
     */
    void ActivateFunctionality(const std::string& _szFunctionality);
    /**
     * @fn void DeactivateFunctionality();
     * @brief Unregisters from previusly activated functionality.
     * @param _szFunctionality Identifier of a functionality.
     * @throw Poco::RuntimeException
     */
    void DeactivateFunctionality(const std::string& _szFunctionality);
    /**
     * @fn void SendEvent(CEventMessage& _Event, const std::string& _szFunctionality = "");
     * @brief Sends event about messenger user.
     * @param _Event Event message.
     * @param _szFunctionality Identifier of a functionality.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void SendEvent(CEventMessage& _Event, const std::string& _szFunctionality = "");
    /**
     * @fn CUser::Ptr GetUser(const std::string& _szTag);
     * @brief Returns a messaging user for an easier interaction.
     * @param _szTag User tag.
     * @return  Messaging user
     */
    CUser::Ptr GetUser(const std::string& _szTag);
    /**
     * @fn void RegisterConnectionListener(Messenger::IConnectionListener* _pConnectionListener);
     * @brief Function used to register for events about network status.
     * @param _pConnectionListener Network connection listener.
     */
    void RegisterConnectionListener(Messenger::IConnectionListener* _pConnectionListener);
    /**
     * @fn void UnregisterConnectionListener();
     * @brief Unregisters network connection listener.
     */
    void UnregisterConnectionListener();
    /**
     * @brief Register event listener to call if there is no listener registered to a specific topic. This can be used for wildcard subscriptions.
     * @param _pListener Event and LWM listener.
     */
    void RegisterDeafultListener(IEventListener* _pListener);
protected:
    friend class CUser;
    /**
     * @fn void RegisterForEvents(IEventListener* _pEventListener, const std::string& _szTag, const std::string& _szFunctionality="");
     * @brief Registers for events arriving from user with a tag,
     * @param _pEventListener Interface to call when a event arrives.
     * @param _szTag User tag.
     * @param _szFunctionality Identifier of a functionality.
     * @throw Poco::RuntimeException
     */
    void RegisterForEvents(IEventListener* _pEventListener, const std::string& _szTag, const std::string& _szFunctionality = "");
    /**
     * @fn void UnregisterFromEvents(IEventListener* _pEventListener, const std::string& _szTag, const std::string& _szFunctionality="");
     * @brief Unregisters from events bound to a messenger user.
     * @param _pEventListener Interface to call when a event arrives.
     * @param _szTag User tag.
     * @param _szFunctionality Identifier of a functionality.
     * @throw Poco::RuntimeException
     */
    void UnregisterFromEvents(IEventListener* _pEventListener, const std::string& _szTag, const std::string& _szFunctionality = "");
    /**
     * @fn void RegisterForLwm(IEventListener* _pLwmListener, const std::string& _szTag);
     * @brief Registers for LWM arriving from user with a tag,
     * @param _pLwmListener Interface to call when a event arrives.
     * @param _szTag User tag.
     * @throw Poco::RuntimeException
     */
    void RegisterForLwm(IEventListener* _pLwmListener, const std::string& _szTag);
    /**
     * @fn void UnregisterFromEvents(IEventListener* _pEventListener, const std::string& _szTag);
     * @brief Unregisters from LWM bound to a messenger user.
     * @param _pLwmListener Interface to call when a LWM arrives.
     * @param _szTag User tag.
     * @throw Poco::RuntimeException
     */
    void UnregisterFromLwm(IEventListener* _pLwmListener, const std::string& _szTag);
    /**
     * @fn void Request(CRequestMessage::Ptr _pRequest, const std::string& _szTag, const std::string& _szFunctionality="");
     * @brief Sends a request to an user identified with a tag.
     * @param _pRequest Request message.
     * @param _szTag User tag.
     * @param _szFunctionality Identifier of a functionality.
     * @throw Poco::RuntimeException
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void Request(CRequestMessage::Ptr _pRequest, const std::string& _szTag, const std::string& _szFunctionality = "");
    /**
     * @fn void CleanRequest(unsigned int _nReqId);
     * @brief Removes a request from the queue.
     * @param _nReqId Request identifier.
     */
    virtual void CleanRequest(unsigned int _nReqId);
    /**
     * @fn virtual void OnMessageArrived(const std::string& _szTopic, const Mqtt::CMessage::Ptr _pMsg);
     * @brief Handles messages arrived from the network.
     * @param _szTopic Message topic.
     * @param _pMsg Message.
     */
    virtual void OnMessageArrived(const std::string& _szTopic, const Mqtt::CMessage::Ptr _pMsg);
private:
    /**
     * @brief Initialize Configure the last will message and subscribe to the response topic.
     */
    void Initialize();
    /**
     * @brief GenerateWillMessage Generate last will message with corresponding status
     * @param _bOnline Indication of the presence
     * @return Generated message
     */
    Mqtt::CMessage::Ptr GenerateStatusMessage(bool _bOnline) const;
    /**
     * @fn void CleanRequests();
     * @brief Cleans all request from the queue. If a request is in the list, failure won't be reported for this request.
     */
    void CleanRequests();
    /**
     * @fn void ProcessEvent(const std::string& _szSenderTag, const Poco::DynamicStruct::Ptr _pPayload);
     * @brief Processes an event message.
     * @param _Msg original MQTT message.
     * @param _szSenderTag User tag of a sender.
     * @param _pPayload Event message.
     * @throw Poco::RuntimeException
     */
    void ProcessEvent(const Mqtt::CMessage& _Msg, const std::string& _szSenderTag, const Poco::DynamicStruct::Ptr _pPayload);
    /**
     * @fn void ProcessLastwill(const std::string& _szSenderTag, const Poco::DynamicStruct::Ptr _pPayload);
     * @brief Processes a lastwill message.
     * @param _Msg original MQTT message.
     * @param _szSenderTag User tag of a sender.
     * @param _pPayload Lastwill message.
     * @throw Poco::RuntimeException
     */
    void ProcessLastwill(const Mqtt::CMessage& _Msg, const std::string& _szSenderTag, const Poco::DynamicStruct::Ptr _pPayload);
    /**
     * @fn void ProcessRequest(const Poco::DynamicStruct::Ptr _pPayload);
     * @brief Processes a request message.
     * @param _Msg original MQTT message.
     * @param _pPayload Request message.
     * @param _szFunctionality Functionality identifier.
     * @throw Poco::RuntimeException
     */
    void ProcessRequest(const Mqtt::CMessage& _Msg, const Poco::DynamicStruct::Ptr _pPayload, const std::string& _szFunctionality = "");
    /**
     * @fn void ProcessFailedRequest(const Poco::DynamicStruct::Ptr _pPayload);
     * @brief Processes a failed request message.
     * @param _pPayload Request message.
     * @throw Poco::RuntimeException
     */
    void ProcessFailedRequest(const Poco::DynamicStruct::Ptr _pPayload);
    /**
     * @fn void ProcessResponse(const Poco::DynamicStruct::Ptr _pPayload);
     * @brief Processes a response message.
     * @param _Msg original MQTT message.
     * @param _pPayload Response message.
     * @throw Poco::RuntimeException
     */
    void ProcessResponse(const Mqtt::CMessage& _Msg, const Poco::DynamicStruct::Ptr _pPayload);
    /**
     * @fn void SendResponse(const std::string& _szReceiverTag, CResponseMessage::Ptr _pResponse);
     * @brief Sends a response to a messaging user.
     * @param _szReceiverTag User tag.
     * @param _Response Response message.
     * @throw Poco::RuntimeException
     * @throw Poco::NullPointerException
     */
    void SendResponse(const std::string& _szReceiverTag, CResponseMessage::Ptr _pResponse);
    /**
     * @fn virtual void OnPublishFailed(int _nCode, const std::string &_szTopic, const Mqtt::CMessage::Ptr _pMsg);
     * @brief Notifies about publish failure.
     * @param _nCode Error code.
     * @param _szTopic Topic.
     * @param _pMsg MQTT message.
     */
    virtual void OnPublishFailed(int _nCode, const std::string& _szTopic, const Mqtt::CMessage::Ptr _pMsg);
    /**
     * @fn static Poco::DynamicStruct::Ptr ParseObject(const std::string& _szObject);
     * @brief Parse a JSON object from a string.
     * @param _szObject JSON string.
     * @return JSON object.
     * @throw Poco::RuntimeException
     */
    Poco::DynamicStruct::Ptr ParseObject(const std::string& _szObject);

    void InspectRetainedFlag(bool _bRetained, const std::string& _szName, const std::string& _szSender);
    void InspectQoS(Mqtt::CMessage::QosLevel _eQos, Mqtt::CMessage::QosLevel _eQosExpected, const std::string& _szName, const std::string& _szSender);

    typedef std::list<IEventListener*> event_listeners_t;
    typedef std::map< std::string, event_listeners_t > event_listener_map_t;
    typedef std::map<unsigned int, CRequestMessage::Ptr> requests_t;

    const std::string m_szUserTag;
    Mqtt::CBuffClient m_ObloClient;
    bool m_bInitialized;

    IResponder* m_pResponder;
    Poco::Mutex m_EvtLwmMutex;
    event_listener_map_t m_EventListeners;
    event_listener_map_t m_LwmListeners;
    Poco::Mutex m_RequestsMutex;
    requests_t m_Requests;
    Poco::RWLock m_ProcessingSync;
    IEventListener* m_pDefaultListener;

    std::string m_szRequestTopic;
    std::string m_szResponseTopic;
    std::string m_szEventTopic;
    std::string m_szStatusTopic;

    const ProtocolVersion m_eProtocolVersion;
};

}
#endif
