/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file MqttClient.h
 * \brief
 * CPP wrapper for PAHO C MQTT client implementation.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_MQTTCLIENT_H
#define OBLO_MESSENGER_MQTTCLIENT_H

extern "C" {
#include "paho/MQTTClient.h"
}

#include "Poco/Mutex.h"
#include "Poco/SharedPtr.h"
#include "Poco/Thread.h"

#include "messenger/Common.h"
#include "messenger/MqttMessage.h"
#include "messenger/MqttConfiguration.h"
#include "messenger/Publishing.h"

namespace Mqtt {

class IDeliveryToken {
public:
    virtual CMessage GetMessage() const=0;
};

class IStatusListener {
public:
    virtual void ConnectionLost(const std::string& _szCause)=0;
    virtual void MessageArrived(const std::string& _szTopic, const CMessage::Ptr _pMsg)=0;
    virtual void DeliveryComplete(const IDeliveryToken& _Token)=0;
    virtual void PublishFailed(int _nCode, const std::string& _szTopic, const CMessage::Ptr _pMsg)=0;
};

class DLLExport CClientImpl {
public:
    typedef Poco::SharedPtr<CClientImpl> Ptr;
    typedef std::vector<std::string> TopicCollection;
    typedef std::vector<int> QosCollection;

    /**
     * @brief MQTT client implementation
     * @param _szServerUri URI of a MQTT broker. Start with <tcp://> for TCP connections, or <ssl://> for secure connections.
     * @param _szClientId Client identifier.
     */
    CClientImpl(const std::string& _szServerUri, const std::string& _szClientId);
    virtual ~CClientImpl();
    virtual void Connect(CConnectConfig::Ptr _pConfig);
    bool IsSessionPresent() const;
    virtual void Disconnect(int _nTimeout=0);
    std::string GetClientId() const;
    bool IsConnected() const;
    virtual void PublishSync(const std::string& _szTopic, const CMessage::Ptr _pMsg);
    virtual void Subscribe(const TopicCollection& _TopicFilters, const QosCollection& _Qos);
    virtual void Subscribe(const std::string& _szTopicFilter, int _nQos);
    virtual void Unsubscribe(const std::string& _szTopicFilter);
    virtual void Unsubscribe(const TopicCollection& _TopicFilters);
    void RegisterListener(IStatusListener* _pListener);
private:
    CClientImpl();
    CClientImpl(const CClientImpl&);
    CClientImpl& operator=(const CClientImpl&);
    static void OnConnectionLost(void* _pContext, char* _szCause);
    static int OnMessageArrived(void* _pContext, char* _szTopicName,
                                int _nTopicLen, MQTTClient_message* _pMsg);
    static void OnDeliveryComplete(void* _pContext, MQTTClient_token _Token);
    IStatusListener* GetStatusListener() const;
    static std::string GetErrorDescription(int _nErrorCode);

    mutable Poco::Mutex m_Mutex;
    MQTTClient m_Cli;
    std::string m_szClientId;
    IStatusListener* m_pStatusListener;
    bool m_bSessionPresent;
};

}

#endif
