/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file MqttConfiguration.h
 * \brief
 * MQTT client configuration classes.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_MQTTCONFIGURATION_H
#define OBLO_MESSENGER_MQTTCONFIGURATION_H

#include "messenger/Common.h"
#include "messenger/MqttMessage.h"

extern "C" {
#include "paho/MQTTClient.h"
}

namespace Mqtt {
/**
 * @class CWillConfig
 * @brief MQTT Last Will Message configuration.
 */
class DLLExport CWillConfig {
public:
    typedef Poco::SharedPtr<CWillConfig> Ptr;
    /**
     * @fn CWillConfig();
     * @brief Default constructor.
     */
    CWillConfig() {}
    /**
     * @fn std::string GetTopic() const;
     * @brief GetTopic
     * @return
     */
    std::string GetTopic() const;
    /**
     * @fn void SetTopic(const std::string& _szTopic);
     * @brief SetTopic
     * @param _szTopic
     */
    void SetTopic(const std::string& _szTopic);
    /**
     * @fn CMessage::Ptr GetMessage() const;
     * @brief GetMessage
     * @return
     */
    CMessage::Ptr GetMessage() const;
    /**
     * @fn void SetMessage(const CMessage::Ptr _pMsg);
     * @brief SetMessage
     * @param _pMsg
     */
    void SetMessage(const CMessage::Ptr _pMsg);
    /**
     * @fn MQTTClient_willOptions* PackConfig() const;
     * @brief Packs configuration to Paho C library option format.
     * @return
     */
    MQTTClient_willOptions* PackConfig() const;
    /**
     * @fn static void FreeConfig(MQTTClient_willOptions* _pConfig);
     * @brief Dealocates memory that is allocated for LWM configuration.
     * @param _pConfig
     */
    static void FreeConfig(MQTTClient_willOptions* _pConfig);
private:
    /**
     * @fn CWillConfig(const CWillConfig&);
     * @brief Forbidden copy constructor.
     */
    CWillConfig(const CWillConfig&);
    /**
     * @fn CWillConfig& operator=(const CWillConfig&);
     * @brief Forbidden assignment operator.
     */
    CWillConfig& operator=(const CWillConfig&);

    std::string m_szTopic;
    CMessage::Ptr m_pMsg;
};
/**
 * @class CSslConfig
 * @brief MQTT SSL/TLS configuration.
 */
class DLLExport CSslConfig {
public:
    typedef Poco::SharedPtr<CSslConfig> Ptr;
    /**
     * @fn CSslConfig();
     * @brief Default constructor.
     */
    CSslConfig();
    /**
     * @fn std::string GetTrustStore() const;
     * @brief Getter for CA crt file path.
     * @return
     */
    std::string GetTrustStore() const;
    /**
     * @fn void SetTrustStore(const std::string& _szTrustStore);
     * @brief Setter for CA crt file path.
     * @param _szTrustStore
     */
    void SetTrustStore(const std::string& _szTrustStore);
    /**
     * @fn std::string GetKeyStore() const;
     * @brief Getter for client's crt file path.
     * @return
     */
    std::string GetKeyStore() const;
    /**
     * @fn void SetKeyStore(const std::string& _szKeyStore);
     * @brief Setter for client's crt file path.
     * @param _szKeyStore
     */
    void SetKeyStore(const std::string& _szKeyStore);
    /**
     * @fn std::string GetPrivateKey() const;
     * @brief Getter for client's key file path.
     * @return
     */
    std::string GetPrivateKey() const;
    /**
     * @fn void SetPrivateKey(const std::string& _szPrivateKey);
     * @brief Setter for client's key file path.
     * @param _szPrivateKey
     */
    void SetPrivateKey(const std::string& _szPrivateKey);
    /**
     * @fn std::string GetPrivateKeyPassword() const;
     * @brief Getter for client's key password.
     * @return
     */
    std::string GetPrivateKeyPassword() const;
    /**
     * @fn void SetPrivateKeyPassword(const std::string& _szPrivateKeyPassword);
     * @brief Setter for client's key password.
     * @param _szPrivateKeyPassword
     */
    void SetPrivateKeyPassword(const std::string& _szPrivateKeyPassword);
    /**
     * @fn bool GetVerifyServer() const;
     * @brief GetVerifyServer
     * @return
     */
    bool GetVerifyServer() const;
    /**
     * @fn void SetVerifyServer(bool _bVerifyServer);
     * @brief Determines if SSL/TLS connections needs to match CA crt file.
     * @param _bVerifyServer
     */
    void SetVerifyServer(bool _bVerifyServer);
    /**
     * @fn MQTTClient_SSLOptions* PackConfig() const;
     * @brief Packs configuration to Paho C library option format.
     * @return
     */
    MQTTClient_SSLOptions* PackConfig() const;
    /**
     * @fn static void FreeConfig(MQTTClient_SSLOptions* _pConfig);
     * @brief Dealocates memory that is allocated for LWM configuration.
     * @param _pConfig
     */
    static void FreeConfig(MQTTClient_SSLOptions* _pConfig);
private:
    /**
     * @fn CSslConfig(const CSslConfig&);
     * @brief Forbidden copy constructor.
     */
    CSslConfig(const CSslConfig&);
    /**
     * @fn CSslConfig& operator=(const CSslConfig&);
     * @brief Forbidden assignment operator.
     */
    CSslConfig& operator=(const CSslConfig&);

    std::string m_szTrustStore;
    std::string m_szKeyStore;
    std::string m_szPrivateKey;
    std::string m_szPrivateKeyPassword;
    bool m_bVerifyServer;
};
/**
 * @class CConnectConfig
 * @brief MQTT connection configuration.
 */
class DLLExport CConnectConfig {
public:
    typedef Poco::SharedPtr<CConnectConfig> Ptr;
    /**
     * @fn CConnectConfig(const std::string& _szUsername="",
                   const std::string& _szPassword="",
                   bool _bCleanSession=true,
                   int _nKeepAliveInterval=60,
                   int _nTimeout=5,
                   int _nRetryinterval=20,
                   bool _bReliable=true);
     * @brief MQTT connection configuration
     * @param _szUsername MQTT username.
     * @param _szPassword MQTT password.
     * @param _bCleanSession MQTT clean session flag.
     * @param _nKeepAliveInterval MQTT keepalive interval.
     * @param _nTimeout MQTT connection timeout.
     * @param _nRetryinterval MQTT operation retry interval.
     * @param _bReliable Paho reliable flag.
     */
    CConnectConfig(const std::string& _szUsername="",
                   const std::string& _szPassword="",
                   bool _bCleanSession=true,
                   int _nKeepAliveInterval=60,
                   int _nTimeout=5,
                   int _nRetryinterval=20,
                   bool _bReliable=false);
    /**
     * @brief GetServerURI
     * @return Configured MQTT server URI
     */
    std::string GetServerURI() const;
    /**
     * @brief SetServerURI
     * @param _szServerURI MQTT server URI
     */
    void SetServerURI(const std::string& _szServerURI);
    /**
     * @fn int GetKeepAliveInterval() const;
     * @brief GetKeepAliveInterval
     * @return
     */
    int GetKeepAliveInterval() const;
    /**
     * @fn void SetKeepAliveInterval(int _nKeepAliveInterval);
     * @brief SetKeepAliveInterval
     * @param _nKeepAliveInterval
     */
    void SetKeepAliveInterval(int _nKeepAliveInterval);
    /**
     * @fn bool GetCleanSession() const;
     * @brief GetCleanSession
     * @return
     */
    bool GetCleanSession() const;
    /**
     * @fn void SetCleanSession(bool _bCleanSession);
     * @brief SetCleanSession
     * @param _bCleanSession
     */
    void SetCleanSession(bool _bCleanSession);
    /**
     * @fn bool IsReliable() const;
     * @brief IsReliable
     * @return
     */
    bool IsReliable() const;
    /**
     * @fn void SetReliable(bool _bReliable);
     * @brief SetReliable
     * @param _bReliable
     */
    void SetReliable(bool _bReliable);
    /**
     * @fn CWillConfig::Ptr GetWill() const;
     * @brief GetWill
     * @return
     */
    CWillConfig::Ptr GetWill() const;
    /**
     * @fn void SetWill(const CWillConfig::Ptr& _pWill);
     * @brief SetWill
     * @param _pWill
     */
    void SetWill(const CWillConfig::Ptr& _pWill);
    /**
     * @fn std::string GetUsername() const;
     * @brief GetUsername
     * @return
     */
    std::string GetUsername() const;
    /**
     * @fn void SetUsername(const std::string& _szUsername);
     * @brief SetUsername
     * @param _szUsername
     */
    void SetUsername(const std::string& _szUsername);
    /**
     * @fn std::string GetPassword() const;
     * @brief GetPassword
     * @return
     */
    std::string GetPassword() const;
    /**
     * @fn void SetPassword(const std::string& _szPassword);
     * @brief SetPassword
     * @param _szPassword
     */
    void SetPassword(const std::string& _szPassword);
    /**
     * @fn int GetTimeout() const;
     * @brief GetTimeout
     * @return
     */
    int GetTimeout() const;
    /**
     * @fn void SetTimeout(int _nTimeout);
     * @brief SetTimeout
     * @param _nTimeout
     */
    void SetTimeout(int _nTimeout);
    /**
     * @fn int GetRetryInterval() const;
     * @brief GetRetryInterval
     * @return
     */
    int GetRetryInterval() const;
    /**
     * @fn void SetRetryInterval(int _nRetryInterval);
     * @brief SetRetryInterval
     * @param _nRetryInterval
     */
    void SetRetryInterval(int _nRetryInterval);
    /**
     * @fn CSslConfig::Ptr GetSsl() const;
     * @brief GetSsl
     * @return
     */
    CSslConfig::Ptr GetSsl() const;
    /**
     * @fn void SetSsl(const CSslConfig::Ptr& _pSsl);
     * @brief SetSsl
     * @param _pSsl
     */
    void SetSsl(const CSslConfig::Ptr& _pSsl);
    /**
     * @fn MQTTClient_connectOptions* PackConfig() const;
     * @brief Packs configuration to Paho C library option format.
     * @return
     */
    MQTTClient_connectOptions* PackConfig() const;
    /**
     * @fn static void FreeConfig(MQTTClient_connectOptions* _pConfig);
     * @brief Dealocates memory that is allocated for LWM configuration.
     * @param _pConfig
     */
    static void FreeConfig(MQTTClient_connectOptions* _pConfig);

private:
    /**
     * @fn CConnectConfig(const CConnectConfig&);
     * @brief Forbidden copy constructor.
     */
    CConnectConfig(const CConnectConfig&);
    /**
     * @fn CConnectConfig& operator=(const CConnectConfig&);
     * @brief Forbidden assignment operator.
     */
    CConnectConfig& operator=(const CConnectConfig&);

    std::string m_szServerURI;
    int m_nKeepAliveInterval;
    bool m_bCleanSession;
    bool m_bReliable;
    CWillConfig::Ptr m_pWill;
    std::string m_szUsername;
    std::string m_szPassword;
    int m_nTimeout;
    int m_nRetryInterval;
    CSslConfig::Ptr m_pSsl;
};

}

#endif
