/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file Network.h
 * \brief
 * Classes related to the network connection.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_NETWORK_H
#define OBLO_MESSENGER_NETWORK_H
#include "Poco/Timer.h"
#include "Poco/Timestamp.h"
#include <Poco/Random.h>

#include "messenger/Common.h"
#include "messenger/Logging.h"

namespace Messenger {
/**
 * @interface IConnectionListener
 * @brief Listener on a connection status.
 */
class IConnectionListener {
public:
    /**
     * @fn virtual void OnConnectionLost(const std::string& _szCause)=0;
     * @brief Callback method which is called when a connection is lost.
     * @param _szCause Connection lost cause.
     */
    virtual void OnConnectionLost(const std::string& _szCause)=0;
};
/**
 * @interface IConnector
 * @brief Interface which is used to initiate a connection.
 */
class IConnector {
public:
    /**
     * @fn virtual void Connect()=0;
     * @brief Callback which is called when there is a need for a connect action.
     */
    virtual void Connect()=0;
};
/**
 * @class CResurector
 * @brief Implementation of a simple reconnection procedure.
 */
class DLLExport CResurector: public IConnectionListener {
public:
    /**
     * @fn CResurector(IConnector& _Connector, unsigned int _nPeriod);
     * @brief Constructor.
     * @param _Connector Interface to use when an action is required.
     * @param _nPeriod Interval for the reconnection procedure (in seconds).
     */
    CResurector(IConnector& _Connector, unsigned int _nInterval=15);
    /**
     * @fn virtual ~CResurector();
     * @brief Destructor.
     */
    virtual ~CResurector();
    /**
     * @brief Set reconnection interval.
     * @param _nInterval Fixed interval in seconds.
     */
    void SetInterval(unsigned int _nInterval);
    /**
     * @brief Set reconnection variable interval. Interval will be generated randomly in range of minimum and maximum.
     * @param _nMinimum - shortest period to wait until next retry,
     * @param _nMaximum - longest period to wait until next retry.
     */
    void SetIntervalRange(unsigned int _nMinimum, unsigned int _nMaximum);
    /**
     * @fn void RegisterLogListener(ILogListener* _pLogListener);
     * @brief RegisterLogListener
     * @param _pLogListener
     */
    void RegisterLogListener(ILogListener* _pLogListener);
    /**
     * @fn virtual void OnConnectionLost(const std::string& _szCause);
     * @brief method which will be called when a connection is broken.
     * @param _szCause
     */
    virtual void OnConnectionLost(const std::string& _szCause);
private:
    /**
     * @fn virtual void OnReconnect(Poco::Timer& timer);
     * @brief OnReconnect
     * @param timer
     */
    virtual void OnReconnect(Poco::Timer& timer);
    /**
     * @brief GetNextPeriod Calculate period of next reconnect
     * @return Next period in milliseconds, 0 if period should remain the same.
     */
    void ScheduleNextRun();
    static unsigned int GetRandom(unsigned int _nMinimum, unsigned int _nMaximum);

    IConnector& m_Connector;
    unsigned int m_nInterval;
    unsigned int m_nMinimum;
    unsigned int m_nMaximum;
    Poco::Timer* m_pTimer;
    Messenger::ILogListener* m_pLogListener;
    Poco::Mutex m_Mutex;
    Poco::Timestamp m_DisconnectionTime;
    static Poco::Random m_RandomGenerator;
};

}

#endif
