/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file NetworkProxy.h
 * \brief
 * Network proxy implementation.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_NETWORKPROXY_H
#define OBLO_MESSENGER_NETWORKPROXY_H

#include "messenger/Common.h"
#include "messenger/MqttClient.h"

#include <Poco/Thread.h>
#include <Poco/ThreadPool.h>
#include <Poco/TaskManager.h>
#include <Poco/TaskNotification.h>
#include <Poco/Condition.h>

namespace Mqtt {
/**
 * @interface IDeliveryListener
 * @brief Delivery listener.
 */
class IDeliveryListener {
public:
    /**
     * @fn virtual void OnMessageArrived(const std::string& _szTopic, const CMessage::Ptr _pMsg)=0;
     * @brief OnMessageArrived
     * @param _szTopic
     * @param _pMsg
     */
    virtual void OnMessageArrived(const std::string& _szTopic, const CMessage::Ptr _pMsg) = 0;
    /**
     * @fn virtual void OnPublishFailed(int _nCode, const std::string& _szTopic, const CMessage::Ptr _pMsg)=0;
     * @brief OnPublishFailed
     * @param _nCode
     * @param _szTopic
     * @param _pMsg
     */
    virtual void OnPublishFailed(int _nCode, const std::string& _szTopic, const CMessage::Ptr _pMsg) = 0;
};
/**
 * @interface IPublishListener
 * @brief The IPublishListener class
 */
class IPublishListener {
public:
    virtual void OnPublishRequired(const std::string& _szTopic, const CMessage::Ptr _pMsg) = 0;
};
/**
 * @interface IConnListener
 * @brief The IConnListener class
 */
class IConnListener {
public:
    /**
     * @fn virtual void OnConnectionLost(const std::string& _szCause)=0;
     * @brief OnConnectionLost
     * @param _szCause
     */
    virtual void OnConnectionLost(const std::string& _szCause) = 0;
};
/**
 * @interface INetworkProxyListener
 * @brief The INetworkProxyListener class
 */
class INetworkProxyListener:
    public IDeliveryListener,
    public IPublishListener,
    public IConnListener {
};
/**
 * @class CMsgDelegate
 * @brief The CMsgDelegate class
 */
class DLLExport CMsgDelegate {
public:
    typedef Poco::SharedPtr<CMsgDelegate> Ptr;

    CMsgDelegate() :
        m_pObject(NULL),
        m_pStub(NULL) {
    }

    template <class T, void (T::*TMethod)(const std::string& topic, const Mqtt::CMessage& msg)>
    static CMsgDelegate::Ptr CreateDelegate(T* _pObject) {
        CMsgDelegate::Ptr d(new CMsgDelegate());
        d->m_pObject = _pObject;
        d->m_pStub = &MethodStub<T, TMethod>;
        return d;
    }
    void operator()(const std::string& topic, const Mqtt::CMessage& msg) const {
        (*m_pStub)(m_pObject, topic, msg);
    }

private:
    typedef void (*stub_type)(void* m_pObject, const std::string& topic, const Mqtt::CMessage& msg);

    void* m_pObject;
    stub_type m_pStub;

    template <class T, void (T::*TMethod)(const std::string& topic, const Mqtt::CMessage& msg)>
    static void MethodStub(void* _pObject, const std::string& topic, const Mqtt::CMessage& msg) {
        T* p = static_cast<T*>(_pObject);
        (p->*TMethod)(topic, msg);
    }
};
/**
 * @class CConnectionRunner
 * @brief The CConnectionRunner class
 */
class DLLExport CConnectionRunner: public Messenger::CSynchronizedRunner {
public:
    /**
     * @fn CConnectionRunner(IConnListener& _ConnDelegate);
     * @brief CConnectionRunner
     * @param _ConnDelegate
     */
    CConnectionRunner(IConnListener& _ConnDelegate);
    /**
     * @fn virtual void Reinitialize();
     * @brief Reinitialize
     */
    virtual void Reinitialize();
    /**
     * @fn void Report(const std::string& _szCause);
     * @brief Report
     * @param _szCause
     */
    void Report(const std::string& _szCause);
protected:
    /**
     * @fn virtual void DoJob();
     * @brief DoJob
     */
    virtual void DoJob();
private:
    static const std::string NO_LOST;
    std::string m_szCause;
    IConnListener& m_ConnDelegate;
};
/**
 * @brief The CRecvMessageDispatcher class
 */
class CRecvMessageTask: public Poco::Task {
public:
    CRecvMessageTask(INetworkProxyListener& _Listener, CRecvHolder::Ptr _pRecvPacket);

    void runTask();
private:
    INetworkProxyListener& m_Listener;
    CRecvHolder::Ptr m_pRecvPacket;
};
/**
 * @class CAsyncNetworkProxy
 * @brief The CAsyncNetworkProxy class
 */
class DLLExport CAsyncNetworkProxy: public IStatusListener,
    public Messenger::IQueueProcessor<CPublishHolder::Ptr>,
    public Messenger::IQueueProcessor<CRecvHolder::Ptr> {
public:
    /**
     * @fn CAsyncNetworkProxy(INetworkProxyListener& _Listener);
     * @brief Constructor.
     * @param _Listener
     * @param _nMaxRecvThreads Set Maximum allowed threads to use for dispatching received messages.
     */
    CAsyncNetworkProxy(INetworkProxyListener& _Listener);
    /**
     * @fn ~CAsyncNetworkProxy();
     * @brief Destructor.
     */
    ~CAsyncNetworkProxy();
    /**
     * @brief ActivateThreadPool
     * Dispatch received messages from threads taken from a thread pool.
     * By default, thread pool is not activated and all received messages will be dispatched from a single thread.
     * @param _nMinThreads Set Minimum threads to run in the pool
     * @param _nMaxRecvThreads Set Maximum allowed threads to use for dispatching
     */
    void ActivateThreadPool(int _nMinThreads, int _nMaxThreads);
    /**
     * @brief DeactivateThreadPool Stop dispatching from thread pool, if previouslly configured
     */
    void DeactivateThreadPool();
    /**
     * @fn void Publish(const std::string& _szTopic, const CMessage::Ptr _pMsg);
     * @brief Publish
     * @param _szTopic
     * @param _pMsg
     */
    void Publish(const std::string& _szTopic, const CMessage::Ptr _pMsg);
private:
    /**
     * @fn virtual void ConnectionLost(const std::string& _szCause);
     * @brief ConnectionLost
     * @param _szCause
     */
    virtual void ConnectionLost(const std::string& _szCause);
    /**
     * @fn virtual void MessageArrived(const std::string& _szTopic, const CMessage::Ptr _pMsg);
     * @brief MessageArrived
     * @param _szTopic
     * @param _pMsg
     */
    virtual void MessageArrived(const std::string& _szTopic, const CMessage::Ptr _pMsg);
    /**
     * @fn virtual void DeliveryComplete(const IDeliveryToken& _Token);
     * @brief DeliveryComplete
     * @param _Token
     */
    virtual void DeliveryComplete(const IDeliveryToken& _Token);
    /**
     * @fn virtual void PublishFailed(int _nCode, const std::string& _szTopic, const CMessage::Ptr _pMsg);
     * @brief PublishFailed
     * @param _nCode
     * @param _szTopic
     * @param _pMsg
     */
    virtual void PublishFailed(int _nCode, const std::string& _szTopic, const CMessage::Ptr _pMsg);
    /**
     * @fn virtual void OnProcess(CPublishHolder::Ptr _Object);
     * @brief OnProcess
     * @param _Object
     */
    virtual void OnProcess(CPublishHolder::Ptr _Object);
    /**
     * @fn virtual void OnProcess(CRecvHolder::Ptr _Object);
     * @brief OnProcess
     * @param _Object
     */
    virtual void OnProcess(CRecvHolder::Ptr _Object);
    /**
     * @brief HandleNotification Receive notification about task status
     * @param _pNotification Notification object
     */
    void HandleNotification(Poco::TaskFinishedNotification* _pNotification);

    INetworkProxyListener& m_Listener;
    Messenger::CQueueRunner<CPublishHolder::Ptr> m_SendingQueue;
    Poco::Thread m_SendingThread;
    Messenger::CQueueRunner<CRecvHolder::Ptr> m_RecvQueue;
    Poco::Thread m_RecvThread;

    Poco::ThreadPool* m_pRecvThreadPool;
    Poco::TaskManager* m_pRecvTaskManager;
    Poco::Mutex m_RecvMutex;
    Poco::Condition m_RecvCondition;

    CConnectionRunner m_Conn;
    Poco::Thread m_ConnThread;
    bool m_bStoped;
};

}

#endif
