/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file Runners.h
 * \brief
 * Synchronized runnables implementation.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_RUNNERS_H
#define OBLO_MESSENGER_RUNNERS_H

#include "Poco/Runnable.h"
#include "Poco/SynchronizedObject.h"
#include "Poco/Mutex.h"
#include "messenger/Common.h"
#include <queue>

namespace Messenger {
/**
 * @class  CSynchronizedRunner
 * @brief Asynchronous abstract class.
 */
class DLLExport CSynchronizedRunner: public Poco::Runnable {
public:
    /**
     * @fn CSynchronizedRunner
     * @brief Constructor.
     */
    CSynchronizedRunner();
    /**
     * @fn virtual ~CSynchronizedRunner();
     * @brief Destructor.
     */
    virtual ~CSynchronizedRunner();
    /**
     * @fn virtual void Reinitialize();
     * @brief Reinitialize
     */
    virtual void Reinitialize();
    /**
     * @fn virtual void Stop();
     * @brief Stop
     */
    virtual void Stop();
protected:
    /**
     * @fn virtual void Pause();
     * @brief Pause
     */
    void Pause();
    /**
     * @fn virtual void Continue();
     * @brief Continue
     */
    void Continue();
    /**
     * @fn virtual void Lock();
     * @brief Lock
     */
    void Lock();
    /**
     * @fn virtual void Unlock();
     * @brief Unlock
     */
    void Unlock();
    /**
     * @fn bool IsRunning() const;
     * @brief IsRunning
     * @return
     */
    bool IsRunning() const;
    /**
     * @fn virtual void DoJob()=0;
     * @brief DoJob
     */
    virtual void DoJob()=0;
private:
    /**
     * @fn virtual void run();
     * @brief run
     */
    virtual void run();
    bool m_bRunning;
    Poco::SynchronizedObject m_Sync;
    Poco::Mutex m_Mutex;
};
/**
 * @interface  IQueueProcessor
 * @brief Interface for an asynchronous processing.
 */
template <class T>
class IQueueProcessor {
public:
    virtual void OnProcess(T _Object)=0;
};
/**
 * @class  CQueueRunner
 * @brief Implementation of an asynchronous processing.
 */
template <class T>
class DLLExport CQueueRunner: public Messenger::CSynchronizedRunner {
public:
    /**
     * @fn CQueueRunner(IQueueProcessor<T>& _Processor);
     * @brief Constructor.
     * @param _Processor
     */
    CQueueRunner(IQueueProcessor<T>& _Processor):
        m_Processor(_Processor) {
    }
    /**
     * @fn virtual ~CQueueRunner();
     * @brief Destructor.
     */
    virtual ~CQueueRunner() {}
    /**
     * @fn void Add(T& _Element);
     * @brief Add
     * @param _Element
     */
    void Add(T _Element) {
        Lock();
        m_Queue.push(_Element);
        Unlock();
        Continue();
    }
protected:
    /**
     * @fn virtual void DoJob();
     * @brief DoJob
     */
    virtual void DoJob() {
        Lock();

        if ( m_Queue.empty() ) {
            Unlock();
            Pause();
        } else {
            while ( !m_Queue.empty() && IsRunning() ) {
                T elm = m_Queue.front();
                m_Queue.pop();
                Unlock();
                m_Processor.OnProcess(elm);
                Lock();
            }

            Unlock();
        }
    }
private:
    typedef std::queue<T> Queue;
    Queue m_Queue;
    IQueueProcessor<T>& m_Processor;
};

}

#endif
