/****************************************************************************
 *  Copyright (c) 2015 RT-RK d.o.o
 *
 *  All Rights Reserved
 *
 *  P R O P R I E T A R Y    &    C O N F I D E N T I A L
 *
 *  -----------------------------------------------------
 *  http://www.rt-rk.com
 *  -----------------------------------------------------
 *
 * \project Oblo Message Broker - messaging library
 * \file User.h
 * \brief
 * Implementation of the messagining user class.
 *
 * @Author Milan Tucic
 *
 * \notes
 *
 * \history
 *****************************************************************************/
#ifndef OBLO_MESSENGER_USER_H
#define OBLO_MESSENGER_USER_H

#include "messenger/Common.h"
#include "messenger/Messages.h"

#include "Poco/Timer.h"

namespace Messenger {

const unsigned int DEFAULT_REQUEST_TIMEOUT = 5;

/**
 * @interface IEventListener
 * @brief Used to report new events from messaging objects.
 * @see CMsgObject
 */
class IEventListener {
public:
    /**
     * @fn virtual void OnEvent(const std::string _szSenderTag, const CEventMessage::Ptr _pEvent) = 0;
     * @brief Called when an event arrives.
     * @param _szSenderTag Sender's tag.
     * @param _pEvent Event message.
     */
    virtual void OnEvent(const std::string& _szSenderTag, const CEventMessage::Ptr _pEvent) = 0;
    /**
     * @fn virtual void OnLastwill(const std::string _szSenderTag, const CLastwillMessage::Ptr _pLastwill) = 0;
     * @brief Called when a lastwill arrives.
     * @param _szSenderTag Sender's tag.
     * @param _pLastwill Lastwill message.
     */
    virtual void OnLastwill(const std::string& _szSenderTag, const CLastwillMessage::Ptr _pLastwill) = 0;
};

class CUser;

/**
 * @class CFunctionality
 * @brief The wrapper for messaging user functionality.
 */
class DLLExport CFunctionality {
public:
    typedef Poco::SharedPtr<CFunctionality> Ptr;
    /**
     * @fn void RegisterForEvents(IEventListener* _pEventListener);
     * @brief Registers for events from user wrapped inside current object. Use NULL to get events to default listener.
     * @param _pEventListener Event listener.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void RegisterForEvents(IEventListener* _pEventListener);
    /**
     * @fn void UnregisterFromEvents(const std::string& _szFunctionality="");
     * @brief Unregisters from events, if previously registered.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     * @throw Poco::NullPointerException
     */
    void UnregisterFromEvents();
    /**
     * @fn const CResponseMessage::Ptr Request(const CRequestMessage& _Request, unsigned int _nTimeout);
     * @brief Sends a synchronous request to an user.
     * @param _Request Request message.
     * @param _nTimeout Time to wait for a response in seconds.
     * @return Response to a request.
     * @throw Poco::TimeoutException
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     * @throw Poco::RangeException
     * @throw Poco::NullPointerException
     */
    const CResponseMessage::Ptr Request(const CRequestMessage& _Request,
                                        unsigned int _nTimeout=DEFAULT_REQUEST_TIMEOUT);
    /**
     * @fn void Request(const CRequestMessage& _Request, const CRequestDelegate::Ptr _pDelegate, unsigned int _nTimeout);
     * @brief Sends an asynchronous request to an user.
     * @param _Request Request message.
     * @param _pDelegate Delegate to call when response arrives.
     * @param _nTimeout Time to wait for a response in seconds.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     * @throw Poco::NullPointerException
     * @throw Poco::RangeException
     * @throw Poco::NullPointerException
     */
    void Request(const CRequestMessage& _Request, const CRequestDelegate::Ptr _pDelegate,
                 unsigned int _nTimeout=DEFAULT_REQUEST_TIMEOUT);
private:
    friend class CUser;
    /**
     * @fn CFunctionality(const std::string& _szName);
     * @brief Functionality constructor.
     * @param _szName Functionality name.
     */
    CFunctionality(const std::string& _szName);
    /**
     * @fn void SetOwner(CUser* _pUser);
     * @brief Setter for messaging user.
     * @param _pUser Messaging user.
     */
    void SetOwner(CUser* _pUser);
    std::string m_szName;
    CUser* m_pUser;
};

class CMessenger;

/**
 * @class CUser
 * @brief Parties included in network communication.
 */
class DLLExport CUser {
public:
    typedef Poco::SharedPtr<CUser> Ptr;
    /**
     * @fn CUser(const std::string& _szTag);
     * @brief Constructor.
     * @param _szTag User's identifier.
     */
    CUser(const std::string& _szTag);
    /**
     * @fn CUser(const CUser& _Obj);
     * @brief Copy constructor.
     * @param _Obj
     */
    CUser(const CUser& _Obj);
    /**
     * @fn virtual ~CUser();
     * @brief Destructor.
     */
    virtual ~CUser();
    /**
     * @fn std::string GetTag() const;
     * @return User's messaging tag.
     */
    std::string GetTag() const;
    /**
     * @fn void RegisterForEvents(IEventListener* _pEventListener);
     * @brief Registers for events from user wrapped inside current object. Use NULL to get events to default listener.
     * @param _pEventListener Event listener.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void RegisterForEvents(IEventListener* _pEventListener);
    /**
     * @fn void UnregisterFromEvents(const std::string& _szFunctionality="");
     * @brief Unregisters from events, if previously registered.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void UnregisterFromEvents();
    /**
     * @fn void RegisterForLwm(IEventListener* _pLwmListener);
     * @brief Registers for lwms from user wrapped inside current object. Use NULL to get events to default listener.
     * @param _pEventListener Event listener.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void RegisterForLwm(IEventListener* _pLwmListener);
    /**
     * @fn void UnregisterFromLwm();
     * @brief Unregisters from LWM, if previously registered.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void UnregisterFromLwm();
    /**
     * @fn const CResponseMessage::Ptr Request(const CRequestMessage& _Request, unsigned int _nTimeout);
     * @brief Sends a synchronous request to an user.
     * @param _Request Request message.
     * @param _nTimeout Time to wait for a response in seconds.
     * @return Response to a request.
     * @throw Poco::TimeoutException
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     * @throw Poco::RangeException
     */
    const CResponseMessage::Ptr Request(const CRequestMessage& _Request, unsigned int _nTimeout=DEFAULT_REQUEST_TIMEOUT);
    /**
     * @fn void Request(const CRequestMessage& _Request, const CRequestDelegate::Ptr _pDelegate, unsigned int _nTimeout);
     * @brief Sends an asynchronous request to an user.
     * @param _Request Request message.
     * @param _pDelegate Delegate to call when response arrives.
     * @param _nTimeout Time to wait for a response in seconds.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     * @throw Poco::NullPointerException
     * @throw Poco::RangeException
     */
    void Request(const CRequestMessage& _Request, const CRequestDelegate::Ptr _pDelegate,
                 unsigned int _nTimeout=DEFAULT_REQUEST_TIMEOUT);
    /**
     * @fn void SendResponse(unsigned int _nRequestId, CResponseMessage::Ptr _pResponse);
     * @brief Sends a response to a messaging user.
     * @param _nRequestId Identifier of a matching request.
     * @param _pResponse message.
     * @throw Poco::RuntimeException
     * @throw Poco::NullPointerException
     */
    void SendResponse(unsigned int _nRequestId, CResponseMessage::Ptr _pResponse);
    /**
     * @fn CFunctionality::Ptr GetFunctionality(const std::string& _szFunctName) const;
     * @brief Getter for an user's functionality.
     * @param _szFunctName Functionality name.
     * @return Pointer to functionality.
     */
    CFunctionality::Ptr GetFunctionality(const std::string& _szFunctName);
protected:
    friend class CMessenger;
    friend class CFunctionality;
    /**
     * @fn void SetMessenger(CMessenger* _pMessenger);
     * @param _pMessenger
     */
    void SetMessenger(CMessenger* _pMessenger);
private:
    /**
     * @fn void RegisterForEvents(IEventListener* _pEventListener, const std::string& _szFunctionality);
     * @brief Registers for events from user wrapped inside current object.
     * @param _pEventListener Event listener.
     * @param _szFunctionality Identifier of an user's functionality.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void RegisterForEvents(IEventListener* _pEventListener, const std::string& _szFunctionality);
    /**
     * @fn void UnregisterFromEvents(const std::string& _szFunctionality);
     * @brief Unregisters from events, if previously registered.
     * @param _szFunctionality Identifier of an user's functionality.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    void UnregisterFromEvents(const std::string& _szFunctionality);
    /**
     * @fn const CResponseMessage::Ptr Request(const CRequestMessage& _Request, const std::string& _szFunctionality,
                                        unsigned int _nTimeout=DEFAULT_REQUEST_TIMEOUT);
     * @brief Sends a synchronous request to an user.
     * @param _Request Request message.
     * @param _szFunctionalityIdentifier of an user's functionality.
     * @param _nTimeout Time to wait for a response in seconds.
     * @return Response to a request.
     * @throw Poco::TimeoutException
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     */
    const CResponseMessage::Ptr Request(const CRequestMessage& _Request, const std::string& _szFunctionality,
                                        unsigned int _nTimeout=DEFAULT_REQUEST_TIMEOUT);
    /**
     * @fnvoid Request(const CRequestMessage& _Request, const CRequestDelegate::Ptr _pDelegate, unsigned int _nTimeout);
     * @brief Sends an asynchronous request to an user.
     * @param _Request Request message.
     * @param _pDelegate Delegate to call when response arrives.
     * @param _szFunctionality Identifier of an user's functionality.
     * @param _nTimeout Time to wait for a response in seconds.
     * @throw Poco::RuntimeException
     * @throw Poco::IllegalStateException
     * @throw Poco::NullPointerException
     */
    void Request(const CRequestMessage& _Request, const CRequestDelegate::Ptr _pDelegate, const std::string& _szFunctionality,
                 unsigned int _nTimeout=DEFAULT_REQUEST_TIMEOUT);

    std::string m_szTag;
    CMessenger* m_pMessenger;
    IEventListener* m_pEventListener;
    IEventListener* m_pLwmListener;
};

}

#endif
